package net.pricefx.common.apibuilder.quote;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import com.fasterxml.jackson.annotation.JsonIgnore;

/**
 * This is a helper object that is used to build up code generated content of a quote.
 * A single QuoteStructure object represents a specified level in the quote folder tree, i.e. it is not representing the full quote if you have multiple subfolder levels
 * <br/>
 * The level is defined by <code>parentId</code> of the structure object
 * 
 *
 */
public class QuoteStructure {

	/**
	 * The list of items to add in this level of the quote
	 */
	@JsonIgnore
	public List<Item> items;
	
	/**
	 * The level of the quote. If parentId == null then the root level is used
	 */
	@JsonIgnore
	public String parentId;
	
	/**
	 * 
	 * A DTO that holds information about the new item (part or folder) to be created.
	 * This is created by the utility methods below. Do not use directly.
	 *
	 */
	public class Item {
		public String labelOrSku;
		public int editabilityStatus;
		public boolean isFolder;
		public List<Item> items;
		public List<Map<String,Object>> inputs;
	}	
	
	/**
	 * Creates a new folder item, but does not add it to the list
	 * @param label The folder label
	 * @return The new folder
	 */
	public Item createFolder(String label) {
		Item i = new Item();
		i.labelOrSku = label;
		i.isFolder = true;
		return i;
	}
	
	/** 
	 * Creates and adds a new folder
	 * @param label The folder label
	 * @return The new folder
	 */
	public Item addFolder(String label) {
		return addFolder(label, null);
	}
	
	/** 
	 * Creates and adds a new folder
	 * @param label The folder label
	 * @return The new folder
	 */
	public Item addFolder(String label, List<Map<String,Object>> inputs) {
		if(items == null) {
			items = new ArrayList<Item>();			
		}
		Item i = createFolder(label);
		i.inputs = inputs;
		items.add(i);
		return i;
	}
	
	/**
	 * Creates a new "part" item but does not add it
	 * @param sku The SKU of the new part
	 * @return The new item
	 */
	public Item createPart(String sku) {
		Item i = new Item();
		i.labelOrSku = sku;
		i.isFolder = false;
		return i;	
	}
	
	/**
	 * Creates and adds a new part
	 * @param sku The SKU of the new part
	 * @return The new item
	 */
	public Item addPart(String sku) {
		return addPart(sku, null);
	}
	
	/**
	 * 
	 * Creates and adds a new part
	 * @param sku The SKU of the new part
	 * @param inputs New inputs of the part
	 * @return The new item
	 */
	public Item addPart(String sku, List<Map<String,Object>> inputs) {
		if(items == null) {
			items = new ArrayList<Item>();			
		}
		Item i = createPart(sku);
		i.inputs = inputs;
		items.add(i);
		return i;
	}	
}
