package net.pricefx.formulaengine;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;


import net.pricefx.common.api.pa.DataType;
import net.pricefx.formulaengine.AbstractProducer.ResultIterator;
import net.pricefx.formulaengine.scripting.Matrix2D;
import net.pricefx.server.protocol.StreamResults;
import org.apache.metamodel.query.Query;
import org.apache.metamodel.schema.Schema;
import org.apache.metamodel.schema.Table;

/**
 * This class is the central starting point for in-memory data table operations.
 * The idea behind it is that you can load bigger chunks of data from PriceAnalyzer or PriceBuilder (or via the List-of-Maps loader really any data) into an in-memory database.
 * From there you can then use normal database semantics (SELECT, JOINS, etc...) to easily pick the data that you really need.
 * Obviously the purpose is to load the bigger data chunk once in the context of e.g. a list calculation then reuse it multiple times or easily combine/join different data structures.
 * <br/><br/>
 * Some notes on the inner workings and the limitations of this feature:
 * <ul>
 * <li> The total number of rows you can store in the in-memory database is limited. Default is 100k rows, but this setting may vary. The number of tables is not limited. The counter is only decreased if you drop entire tables.</li>
 * <li> The in-memory database hinges at the variable you get back from <code>api.getTableContext()</code>. So if that variable goes out of scope the database is lost. Hence for reusing put it into <code>api.global</code> for example.</li>
 * <li> A database in api.global will NOT survive a variable pass-on between two passes of a list calculation (e.g. from first pass to dirty pass calculation)
 * <li> In a distributed calculation every calculation thread has its own database. These databases are not shared in any way.
 * <li> You can easily convert results of a query of the in-memory databases tables into a ResultMatrix. @see net.pricefx.server.util.Matrix2D#toResultMatrix
 * </ul>
 * 
 * <br/><br/>
 * Some example code:
 * <br/><br/>
 * <pre>
 * {@code
 * def tableContext = api.getTableContext()
 * 
 * def fields1 = ["ID":DataType.LONG,
 *   			  "Name":DataType.STRING,
 *                "Value":DataType.STRING]
 * 
 * def fields2 = ["Name":DataType.STRING,
 *                "Multiplier":DataType.NUMBER]
 * 
 * def values1 = [
 * 				   ["ID":1, "Name":"A", "Value":"1"],
 *                 ["ID":2, "Name":"B", "Value":"2"],
 *                 ["ID":3, "Name":"C", "Value":"3"]
 *               ]
 * 
 * def values2 = [
 * 				   ["Name":"A", "Multiplier":0.5],
 *                 ["Name":"B", "Multiplier":2],
 *                 ["Name":"C", "Multiplier":5]
 *               ]
 * tableContext.createTable("Table1", fields1, null)
 * tableContext.loadRows("Table1",values1)
 * 
 * tableContext.createTable("Table2", fields2, null)
 * tableContext.loadRows("Table2",values2)
 * 
 * api.trace("Tables",null,tableContext.printTablesDefinitions("%"))
 * api.trace("Tables 1 Count",null,tableContext.count("Table1"))
 * api.trace("Tables 2 Count",null,tableContext.count("Table2"))
 * 
 * return tableContext.executeQuery("SELECT t1.Name, Value, Multiplier FROM Table1 t1 JOIN Table2 t2 ON (t1.Name = t2.Name)").toResultMatrix()
 * }
 * </pre>
 *
 */
public interface TableContext {

	
	public void close() throws SQLException;

	/**
	 * Returns the total row count over all used in-memory tables. This value is used to determine if the maximum row count for storage is exceeded or not
	 * @return The row count
	 */
	public int getTotalRows();

	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Table management
	//
	
	/**
	 * Drops an existing in-memory table.
	 * Note:
	 * Only by calling this method (and not an executeUpdate query with a DROP statement) the internal row counter is decreased.
	 * 
	 * @param tableName The in-memory table name to drop
	 * @throws SQLException
	 */
	public void dropTable(final String tableName) throws SQLException;

	/**
	 * Creates a new in-memory table - eventually dropping an already existing table with same name
	 * 
	 * @param tableName The in-memory table name
	 * @param input Accepts two types: String and <code>DatamartQueryResultImpl</code>. If a string is passed it needs to be a valid SELECT sql clause based on an already existing in-memory database.
	 * Basically a <code>"CREATE TABLE x AS SELECT ..."</code> command is called. In case of <code>DatamartQueryResultImpl</code> the method behaves as createTableFromQueryResult
	 * @throws SQLException
	 */
	public void createTable(final String tableName, final Object input) throws SQLException;


	/**
	 * Creates a new in-memory table - eventually dropping an already existing table with same name - based on the query result fields.
	 * It then immediately populates the in-memory table with the result data.
	 * 
	 * @param tableName The in-memory table name
	 * @param queryResult The result set of a datamart query
	 * @throws SQLException
	 */
	public void createTableFromQueryResult(final String tableName, final DatamartQueryResult queryResult) throws SQLException;
	
	/**
	 * Creates a new in-memory table - eventually dropping an already existing table with same name - based on the query definition fields.
	 * No data is loaded, nor the datamart query executed
	 * 
	 * @param tableName The in-memory table name
	 * @param query
	 */
	public void createTableFromQuery(final String tableName, final DatamartContext.Query query);
	
	/**
	 * Creates a new in-memory table - eventually dropping an already existing table with same name - based on the product master meta data.
	 * No data is loaded yet.
	 * 
	 * @param tableName The in-memory table name
	 * @param fields The fields to pick form the product master, or null for all fields
	 * @throws SQLException
	 */
	public void createTableFromProductMaster(final String tableName, final Collection<String> fields) throws SQLException;
	/**
	 * Creates a new in-memory table - eventually dropping an already existing table with same name - based on the customer master meta data.
	 * No data is loaded yet.
	 * 
	 * @param tableName The in-memory table name
	 * @param fields The fields to pick form the customer master, or null for all fields
	 * @throws SQLException
	 */	
	public void createTableFromCustomerMaster(final String tableName, final Collection<String> fields) throws SQLException;
	
	/**
	 * Creates a new in-memory table - eventually dropping an already existing table with same name - based on the product master extension meta data.
	 * No data is loaded yet.
	 * 
	 * @param tableName The in-memory table name
	 * @param pxName The extension name
	 * @param fields The fields to pick form the product master extension, or null for all fields
	 * @throws SQLException
	 */	
	public void createTableFromProductExtension(final String tableName, final String pxName, final Collection<String> fields) throws SQLException;

	/**
	 * Creates a new in-memory table - eventually dropping an already existing table with same name - based on the customer master extension meta data.
	 * No data is loaded yet.
	 * 
	 * @param tableName The in-memory table name
	 * @param cxName The extension name
	 * @param fields The fields to pick form the customer master extension, or null for all fields
	 * @throws SQLException
	 */		
	public void createTableFromCustomerExtension(final String tableName, final String cxName, final Collection<String> fields) throws SQLException;

	/**
	 * Creates a new in-memory table - eventually dropping an already existing table with same name - based on a lookup table/ pricing parameter table
	 * 
	 * @param tableName The in-memory table name
	 * @param lookupTableId The id of the lookup table
	 * @param fields The fields to pick form the customer master extension, or null for all fields
	 * @throws SQLException
	 */
	public void createTableFromLookupTable(final String tableName, final Long lookupTableId, final Collection<String> fields) throws SQLException;

	/**
	 * Creates a new in-memory table - eventually dropping an already existing table with same name - based on the passed field list
	 * No data is loaded yet.
	 * 
	 * @param tableName The in-memory table name
	 * @param fields Mandatory. Map of fieldName => DataType of the table. Supported DataType constants are: STRING, NUMBER, DATE, BOOLEAN, DATETIME, LONG
	 * @param indexFields Optional, can be null. List of fieldNames for which an index should be created
	 */
	public void createTable(final String tableName, Map<String,DataType> fields, List<String> indexFields) throws SQLException;

	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Select / Updates
	//
	

	/**
	 * Get the row count of a in-memory table
	 * @param tableName The in-memory table
	 * @return The total row count of that table
	 */
	public int count(final String tableName);
	
	/**
	 * Execute non-select queries like UPDATE,INSERT, etc
	 * @param sql A valid SQL statement
	 * @return The row count of affected rows or 0
	 * @throws SQLException
	 */
	public int executeUpdate(final String sql) throws SQLException;
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Public helpers
	//
	/**
	 * Returns a debug string that contains the complete schema definition for the temporary in-memory database.
	 * Useful for developing and debugging. Do NOT use in a real production logic (i.e. comment it outonce done with debugging)
	 * 
	 * @param tablePattern a table name pattern; must match the table name as it is stored in the database. You can use wildcards like %
	 * @return The formatted schema description
	 * @throws SQLException
	 */
	public String printTablesDefinitions(String tablePattern) throws SQLException;


	/**
	 * Converts a Price f(x) datatype into a in-memory database datatype.
	 * @param dataType
	 * @return
	 */
	public String toDBDataType( final DataType dataType);
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Private helpers
	//

    public Schema getDefaultSchema();

	public Table[] getTables();
	
	public Table getTable(final String tableName);

	public Query newQuery(final Object table) throws SQLException;

	public Query newQuery(final String tableName) throws SQLException;

	public Query parseQuery(final String queryString);

    public Matrix2D executeQuery(final String queryString) throws SQLException;
    
    public Matrix2D executeQuery(final Object query) throws SQLException;
    
    
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Data loading
	//    

    /**
     * Load data into an existing in-memory table. This method can be used in combination with <code>api.stream</code> like:
	 * <pre>
	 * {@code
	 * def pxResults = api.stream("PX", null, Filter.equal("name", "MySamplePX"))
	 * tableContext.loadRows("PX_T1", pxResults)
	 * }
	 * </pre>
     * 
     * @param tableName The target in-memory table
     * @param stream The result of an api.stream call
     * @throws SQLException
     */
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public void loadRows(final String tableName, final ResultIterator stream) throws SQLException;


	@SuppressWarnings({ "unchecked", "rawtypes" })
	public void loadRows(final String tableName, final StreamResults results) throws SQLException;
	
	/**
	 * Load data into an existing in-memory table. This method can be used in combination with <code>api.find</code>
	 * <pre>
	 * {@code
	 * def pxResults = api.find("PX", null, Filter.equal("name", "MySamplePX"))
	 * tableContext.loadRows("PX_T1", pxResults)
	 * }
	 * </pre>
	 * @param tableName The target in-memory table
	 * @param rowsAsListOfMaps The result of an api.find call or any custom list with map elements
	 * @throws SQLException
	 */
	public void loadRows(final String tableName, final List<Map<String,Object>> rowsAsListOfMaps) throws SQLException;

}