package net.pricefx.formulaengine.scripting.inputbuilder;

import net.pricefx.common.api.InputType;
import net.pricefx.common.api.PriceFxInterface;
import net.pricefx.lmformulaengine.LmFormulaEngineContext;

import java.math.BigDecimal;


/**
 * Input Builder Factory class as a convenient way to create inputs (user entries, context parameters).
 * <p>
 * Overview of various input types can be found in article
 * <a href="https://pricefx.atlassian.net/wiki/display/UNITY/Input+Functions">Input Functions</a>.
 * </p>
 * <h2>Example</h2>
 * <blockquote>
 * <pre>def nameInput = api.inputBuilderFactory()
 *         .createTextUserEntry("Name")
 *         .buildContextParameter()
 * </pre>
 * </blockquote>
 *
 * @see PublicGroovyAPI#inputBuilderFactory()
 * @see net.pricefx.server.dto.calculation.ContextParameter
 * @see InputType
 */
public class InputBuilderFactory {
    private final InputBuilderContext builderContext;

    /**
     * Use <code>api.inputBuilderFactory()</code> {@link PublicGroovyAPI#inputBuilderFactory()} to obtain instance of this class from Groovy API.
     * @see PublicGroovyAPI#inputBuilderFactory()
     */
    public InputBuilderFactory(LmFormulaEngineContext context, Object callback) {
        builderContext = new InputBuilderContext(context,callback);
    }

    /**
     * Creates new <code>UserEntry</code> {@link BigDecimal} input (context parameter).
     * <p><b>Side effect</b>: During the Syntax Check mode, this function triggers creation of the input parameter.
     *
     * <ul>
     * <li>When the user-entered number is a whole number (integer),
     * the value returned is java.lang.Integer.</li>
     * <li>When the user-entered number is a decimal number,
     * the value returned is java.math.BigDecimal.</li>
     * <li>If you always need the result as a BigDecimal, you can cast it,
     * e.g. api.decimalUserEntry("number") as BigDecimal.</li>
     * </ul>
     *
     * @param paramName Name of the user input parameter. For the names use the "camel case" (no spaces and capital letters).
     *                     This value is not meant as a label of the input field, but rather a name which should stay
     *                     stable for a long time. Remember, the input fields are stored e.g. on every single
     *                     Quote line item and are created only once when adding the new line. If you need to modify
     *                     the label (or other properties) of the input parameter, you can use getParameter(String).
     *
     * @return User entry input builder.
     * @see PublicGroovyAPI#userEntry(String)
     * @see SimpleInputBuilder
     */
    public SimpleInputBuilder<BigDecimal> createUserEntry(String paramName) {
        return new SimpleInputBuilder<BigDecimal>(builderContext, paramName, InputType.USERENTRY, null);
    }

    /**
     * Creates new <code>IntegerUserEntry</code> {@link Integer} input (context parameter).
     * <p><b>Side effect</b>: During the Syntax Check mode, this function triggers creation of the input parameter
     * that lets the user enter an integer value. The input is rendered as a numeric field.
     *
     * @param paramName Name of the user input parameter. Use "camel case" for the names (no spaces and capital letters).
     *                     This value is not meant as a label of the input field, but rather a name which should stay
     *                     stable for a long time. Remember, the input fields are stored e.g. on every single
     *                     Quote line item and are created only once, when adding the new line. If you need to modify
     *                     the label (or other properties) of the input parameter, you can use getParameter(String).
     *
     * @return Integer user entry builder.
     * @see PublicGroovyAPI#integerUserEntry(String)
     * @see SimpleInputBuilder
     */
    public SimpleInputBuilder<Integer> createIntegerUserEntry(String paramName) {
        return new SimpleInputBuilder<Integer>(builderContext, paramName, InputType.INTEGERUSERENTRY, null);
    }

    /**
     * Creates new <code>BooleanUserEntry</code> {@link Boolean} input (context parameter).
     * Rendered as a checkbox.
     *
     * @param paramName Name of the user entry parameter.
     * @return Boolean user entry builder.
     * @see PublicGroovyAPI#booleanUserEntry(String)
     * @see SimpleInputBuilder
     */
    public SimpleInputBuilder<Boolean> createBooleanUserEntry(String paramName) {
        return new SimpleInputBuilder<Boolean>(builderContext, paramName, InputType.BOOLEANUSERENTRY, null);
    }

    /**
     * Creates new <code>DateUserEntry</code> {@link java.util.Date} input (context parameter).
     * Rendered as a date picker field that lets the user enter a date value.
     *
     * @param paramName Name of the user entry parameter.
     * @return Date user entry builder.
     * @see PublicGroovyAPI#dateUserEntry(String)
     * @see SimpleInputBuilder
     */
    public SimpleInputBuilder<String> createDateUserEntry(String paramName) {
        return new SimpleInputBuilder<String>(builderContext, paramName, InputType.DATEUSERENTRY, null);
    }

    /**
     * Creates new <code>DateRangeUserEntry</code> input (context parameter).Creates an input parameter 
     * Rendered as two date picker fields that let the user enter a date range (start and end date).
     * <p>
     * Note: This is available in Unity UI only.
     *
     * @param paramName Name of the user entry parameter.
     * @return Date range user entry builder.
     * @see PublicGroovyAPI#dateRangeUserEntry(String)
     * @see SimpleInputBuilder
     */
    public DateRangeInputBuilder createDateRangeUserEntry(String paramName) {
        return new DateRangeInputBuilder(builderContext, paramName);
    }

    /**
     * Creates new <code>TimeUserEntry</code> input (context parameter).
     * Rendered as a time picker field that lets the user enter a time value in the format "HH:mm" (e.g. "23:59").
     *
     * @param paramName Name of the user entry parameter.
     * @return Time user entry builder.
     * @see PublicGroovyAPI#timeUserEntry(String)
     * @see SimpleInputBuilder
     */
    public SimpleInputBuilder<String> createTimeUserEntry(String paramName) {
        return new SimpleInputBuilder<String>(builderContext, paramName, InputType.TIMEUSERENTRY, null);
    }

    /**
     * Creates new <code>DateTimeUserEntry</code> input (context parameter).
     * Rendered as a datetime picker field that lets the user enter a date time value in the format "dd/MM/yyyy HH:mm" (e.g. "23/01/2019 10:30").
     * <p>
     * Note: The format implicitly uses the GMT/UTC time zone and so the returned value may
     * differ from what was entered on the client side (as the client time may be in a different time zone).
     *
     * @see #parseDate(String, String)
     * @param paramName Name of the user entry parameter.
     * @return DateTime user entry builder.
     * @see PublicGroovyAPI#dateTimeUserEntry(String)
     * @see InputType#DATETIMEUSERENTRY
     * @see SimpleInputBuilder
     */
    public SimpleInputBuilder<String> createDateTimeUserEntry(String paramName) {
        return new SimpleInputBuilder<String>(builderContext, paramName, InputType.DATETIMEUSERENTRY,null);
    }


    /**
     * Creates new radio input (context parameter).
     * User can select one value from many by clicking a radio button.
     * @param paramName Name of the parameter.
     * @return Radio user entry builder.
     * @see InputType#RADIO
     * @see OptionInputBuilder
     */
    public OptionInputBuilder createRadioEntry(String paramName) {
        return new OptionInputBuilder(builderContext, InputType.RADIO, paramName);
    }

    /**
     * Creates new hidden input (context parameter).
     * This type of input is not visible to the user.
     * @param paramName Name of the parameter.
     * @return Hidden entry builder.
     * @see InputType#HIDDEN
     * @see SimpleInputBuilder
     */
    public SimpleInputBuilder<Object> createHiddenEntry(String paramName) {
        return new SimpleInputBuilder<Object>(builderContext, paramName, InputType.HIDDEN, null);
    }

    /**
     * Creates new <code>StringUserEntry</code> input (context parameter).
     * Returns a text value entered by the user in the input parameter.
     * <p><b>Side effect</b>: During the Syntax Check mode, this function triggers creation of the single-text input parameter.
     *
     * @param paramName Name of the user entry parameter.
     * @return String user entry builder.
     * @see PublicGroovyAPI#stringUserEntry(String)
     * @see InputType#STRINGUSERENTRY
     * @see StringInputBuilder
     */
    public StringInputBuilder<String> createStringUserEntry(String paramName) {
        return new StringInputBuilder<String>(builderContext, paramName);
    }

    /**
     * Creates new <code>TextUserEntry</code> {@link InputType#TEXTUSERENTRY} input (context parameter).
     * Rendered as a multi-line text box that lets the user enter a longer text value.
     *
     * @param entryName Name of the user entry parameter.
     * @return Text user entry builder.
     * @see PublicGroovyAPI#textUserEntry(String)
     * @see PublicGroovyAPI#textUserEntry(String, String)
     * @see InputType#TEXTUSERENTRY
     * @see SimpleInputBuilder
     */
    public SimpleInputBuilder<String> createTextUserEntry(String paramName) {
        return new SimpleInputBuilder<String>(builderContext, paramName, InputType.TEXTUSERENTRY, null);
    }

    /**
     * Creates new <code>UserEntry</code> {@link InputType#USERENTRY} input (context parameter).
     * It is rendered as a user picker dialog / select list.
     *
     * @param paramName Name of the entry parameter.
     * @return User entry input builder.
     * @see PublicGroovyAPI#anyUser(String)
     * @see InputType#USERENTRY
     * @see SimpleInputBuilder
     */
    public SimpleInputBuilder<String> createAnyUserEntry(String paramName) {
        return new SimpleInputBuilder<String>(builderContext, paramName, InputType.USER, null);
    }

    /**
     * Creates new <code>AnyUserEntry</code> {@link InputType#USERENTRY} input (context parameter).
     * Searches for a record in the price parameter table named {@code parameterName}
     * where the column 'name' matches the value entered by the user (in a drop-down input parameter) and returns
     * the value from the column 'value'.
     * <p><b>Side effect</b>: During the Syntax Check mode this function creates a new input parameter (drop-down)
     * whose values are taken from the column "Name" of the given Price Parameter.
     * @param paramName Name of the price parameter table to look up.
     * @return User entry input builder.
     * @see PublicGroovyAPI#vLookup(String)
     * @see PublicGroovyAPI#vLookup(String, String, String, String, String, String, String, String)
     * @see InputType#LOOKUP
     * @see InputType#MATRIXLOOKUP
     * @see VLookupBuilder
     */
    public VLookupBuilder createVLookup(String paramName) {
        return new VLookupBuilder(builderContext, paramName);
    }

    /**
     * Creates new <code>ProductEntry</code> {@link InputType#PRODUCT} input (context parameter).
     * This is an empty product entry object that must be filled with input parameters by calling createParameter on the object.
     *
     * @return Product entry builder.
     * @see PublicGroovyAPI#product()
     * @see InputType#PRODUCT
     * @see ProductInputBuilder
     */
    public ProductInputBuilder createProductEntry() {
        return new ProductInputBuilder(builderContext, PriceFxInterface.CONTEXTPARAM_NAME_PRODUCTSKU);
    }

    /**
     * Creates new <code>ProductEntry</code> {@link InputType#PRODUCT} input (context parameter).
     * This is commonly used in PromotionManager and RebateManager.
     * It is rendered as a product group picker widget.
     *
     * @param paramName Name of the user entry parameter.
     * @return Product entry builder.
     * @see PublicGroovyAPI#product(String)
     * @see InputType#PRODUCT
     * @see ProductInputBuilder
     */
    public ProductInputBuilder createProductEntry(String paramName) {
        return new ProductInputBuilder(builderContext, paramName);
    }

    /**
     * Creates new <code>ProductGroupEntry</code> {@link InputType#PRODUCTGROUP} input (context parameter).
     * This is an empty product group entry object. Must be filled with input parameters by calling createParameter on the object.
     *
     * @return Product group entry builder.
     * @see PublicGroovyAPI#productGroupEntry(String...)
     * @see PublicGroovyAPI#productGroupEntry(String, String, Object)
     * @see InputType#PRODUCTGROUP
     * @see PCGroupInputBuilder
     */
    public PCGroupInputBuilder createProductGroupEntry() {
        return new PCGroupInputBuilder(builderContext, InputType.PRODUCTGROUP, PriceFxInterface.CONTEXTPARAM_NAME_PRODUCTGROUP);
    }

    /**
     * Creates new <code>ProductGroupEntry</code> {@link InputType#PRODUCTGROUP} input (context parameter).
     * Creates an input parameter that lets the user select a group of products.
     * This is commonly used in PromotionManager and RebateManager.
     * It is rendered as a product group picker widget.
     *
     * @param paramName Name of the user entry parameter.
     * @return Product group entry builder.
     * The input parameter returns the user selection as a Map:
     * <pre>
     * [
     *   "productFieldName"  : "attribute3",
     *   "productFieldLabel" : "Business Unit",
     *   "productFieldValue" : "MeatBall"
     * ]
     * </pre>
     * @see PublicGroovyAPI#productGroupEntry(String...)
     * @see PublicGroovyAPI#productGroupEntry(String, String, Object)
     * @see InputType#PRODUCTGROUP
     * @see PCGroupInputBuilder
     */
    public PCGroupInputBuilder createProductGroupEntry(String paramName) {
        return new PCGroupInputBuilder(builderContext, InputType.PRODUCTGROUP, paramName);
    }

    /**
     * Creates new <code>CustomerEntry</code> {@link InputType#CUSTOMER} input (context parameter).
     * This is an empty customer entry object. Must be filled with input parameters by calling createParameter on the object.
     *
     * @return Customer entry builder.
     * @see PublicGroovyAPI#customer(String)
     * @see InputType#CUSTOMER
     * @see CustomerInputBuilder
     */
    public CustomerInputBuilder createCustomerEntry() {
        return new CustomerInputBuilder(builderContext, PriceFxInterface.CONTEXTPARAM_NAME_CUSTOMER);
    }

    /**
     * Creates new <code>CustomerEntry</code> {@link InputType#CUSTOMER} input (context parameter).
     * This is commonly used in PromotionManager and RebateManager.
     * It is rendered as a customer picker widget.
     *
     * @param paramName Name of the user entry parameter.
     * @return Customer entry builder.
     * @see PublicGroovyAPI#customer(String, String)
     * @see PublicGroovyAPI#customer(String, String, boolean, String, Object)
     * @see InputType#CUSTOMER
     * @see CustomerInputBuilder
     */
    public CustomerInputBuilder createCustomerEntry(String paramName) {
        return new CustomerInputBuilder(builderContext, paramName);
    }

    /**
     * Creates new <code>CustomerGroupEntry</code> {@link InputType#CUSTOMERGROUP} input (context parameter).
     * This is an empty customer group entry object. Must be filled with input parameters by calling createParameter on the object.
     *
     * @return Customer group entry builder.
     * @see PublicGroovyAPI#customerGroupEntry(String...)
     * @see InputType#CUSTOMERGROUP
     * @see PCGroupInputBuilder
     */
    public PCGroupInputBuilder createCustomerGroupEntry() {
        return new PCGroupInputBuilder(builderContext, InputType.CUSTOMERGROUP, PriceFxInterface.CONTEXTPARAM_NAME_CUSTOMERGROUP);
    }

    /**
     * Creates new <code>CustomerGroupEntry</code> {@link InputType#CUSTOMERGROUP} input (context parameter).
     * This input parameter lets the user select a group of customers and returns the selection as a Map.
     * This is commonly used in PromotionManager and RebateManager.
     * It is rendered as a customer group picker widget.
     *
     * @param paramName Name of the user entry parameter.
     * @return Customer group input builder.
     * <pre>
     * [
     *   "customerFieldName" : "attribute3",
     *   "customerFieldLabel" : "Customer Type",
     *   "customerFieldValue" : "Restaurant"
     * ]
     * </pre>
     * @see PublicGroovyAPI#customerGroupEntry(String, String, Object)
     * @see InputType#CUSTOMERGROUP
     * @see PCGroupInputBuilder
     */
    public PCGroupInputBuilder createCustomerGroupEntry(String paramName) {
        return new PCGroupInputBuilder(builderContext, InputType.CUSTOMERGROUP, paramName);
    }

    /**
     * Creates new <code>OptionEntry</code> {@link InputType#OPTION} input (context parameter).
     * The input returns a value selected by the user from a list displayed in drop-down input parameter.
     * <p><b>Side effect</b>: During the Syntax Check mode, this function triggers creation of a drop-down
     * input parameter with the given list of options.
     *
     * <h2>Example:</h2>
     * <blockquote>
     *     <pre>
     *         api.inputBuilderFactory
     *                 .createOptionEntry("entryName")
     *                 .setLabel("Entry Label")
     *                 .addOption("name1")
     *                 .addOption("name2")
     *                 .setLabels(["name1": "Label1", "name2": "Label2"])
     *                 .buildContextParameter()
     *     </pre>
     * </blockquote>
     *
     * @param paramName Name of the user entry parameter.
     * @param options Arbitrary number of options that should be displayed in the drop-down list.
     * @return Option input builder.
     * @see PublicGroovyAPI#option(String, Object...)
     * @see PublicGroovyAPI#option(String, List, Map)
     * @see OptionInputBuilder
     */
    public OptionInputBuilder createOptionEntry(String paramName) {
        return new OptionInputBuilder(builderContext, InputType.OPTION, paramName);
    }

    /**
     * Creates new <code>OptionsEntry</code> {@link InputType#OPTIONS} input (context parameter).
     * The input parameter lets the user select multiple values from predefined options.
     * It is rendered as a drop-down list of possible options. Each selected value appears at the top.
     * <h2>Example:</h2>
     * <blockquote>
     *     <pre>
     *         api.inputBuilderFactory
     *                 .createOptionsEntry("entryName")
     *                 .setLabel("Entry Label")
     *                 .addOption("name1")
     *                 .addOption("name2")
     *                 .setLabels(["name1": "Label1", "name2": "Label2"])
     *                 .buildContextParameter()
     *     </pre>
     * </blockquote>
     *
     * @param paramName Name of the user entry parameter.
     * @param options Arbitrary number of options that should be displayed in the drop-down list.
     * @return Option input builder.
     * @see PublicGroovyAPI#options(String, Object...)
     * @see PublicGroovyAPI#options(String, List, Map)
     * @see OptionInputBuilder
     */
    public OptionInputBuilder createOptionsEntry(String paramName) {
        return new OptionInputBuilder(builderContext, InputType.OPTIONS, paramName);
    }

    /**
     * Creates new <code>InputMatrix</code> {@link InputType#INPUTMATRIX} input (context parameter).
     * This input parameter that lets the user select a matrix of values.
     * It renders as a grid-style input widget with the specified columns.
     * The result will be a list (= rows) of maps (= row's columns). The map attribute names are the column names.
     * <p>
     * Make sure the column names are valid JSON identifiers (spaces are ok but try to avoid special chars).
     *
     * @param inputMatrixName Name of the parameter.
     * @param columnNames     Column names/labels.
     * @return Input matrix input builder.
     * @see PublicGroovyAPI#inputMatrix(String, String...)
     * @see PublicGroovyAPI#inputMatrix(String, Map, String...)
     * @see InputMatrixInputBuilder
     */
    public InputMatrixInputBuilder createInputMatrix(String paramName) {
        return new InputMatrixInputBuilder(builderContext, paramName);
    }

    /**
     * Creates new <code>FilterBuilderUserEntry</code> {@link InputType#FILTERBUILDER} input (context parameter).
     * This input parameter lets the user build a filter for a table of objects of the type defined by the {@code typeCode} parameter.
     * <p>
     * It is rendered as a filter builder widget allowing to construct a filter (e.g. to define filter for Products, set the {@code typeCode} to {@code P}).
     * <p>
     * The filter object can be used in other functions, like {@link #find(String, Filter...)}, {@link #stream(String, String, Filter...)}, {@link #count(String, Filter...)} etc.
     *
     * @param paramName Name of the user entry parameter.
     * @param typeCode  Type code string of the type of the object for which the filter is set. Currently, the supported type codes are "P" (Product data) and "C" (Customer data).
     * @return Filter input builder.
     * @see PublicGroovyAPI#filterBuilderUserEntry(String, String)
     * @see FilterBuilderInputBuilder
     */
    public FilterBuilderInputBuilder createFilterBuilder(String paramName, String typeCode) {
        return new FilterBuilderInputBuilder(builderContext, paramName, typeCode);
    }

    /**
     * Creates new <code>ConfiguratorEntry</code> {@link InputType#FILTERBUILDER} input (context parameter).
     * @param paramName name of input
     * @return <code>ConfiguratorInputBuilder</code> to set additional options
     * @see PublicGroovyAPI#createConfiguratorEntry()
     * @see PublicGroovyAPI#createConfiguratorEntry(InputType, String, Date)
     * @see ConfiguratorInputBuilder
     */
    public ConfiguratorInputBuilder createFilterBuilder(String paramName, String formulaName, boolean isInline) {
        return new ConfiguratorInputBuilder(builderContext, paramName, formulaName, isInline);
    }

    /**
     * Creates new <code>DataMartFilterBuilderUserEntry</code> {@link InputType#DMFILTERBUILDER} input (context parameter).
     * This input parameter lets the user build a filter for a given Datamart.
     * <p>
     * The filter can be used in {@link net.pricefx.formulaengine.DatamartContext.Query#where(Filter...)}
     * of the {@link ISandboxAPI#datamartQuery(Object...)}.
     * <p>
     * It will render a FilterBuilder widget entry allowing to construct a filter object filtering on the specified Data Source.
     * <p>
     * Example:
     * <blockquote>
     * <pre>
     * def filter1 = datamartFilterBuilderUserEntry("ProductGroup", "Transactions_DM")
     *
     * def dimFilterList = ["CustomerId", "ProductId"]
     * def filter2 = datamartFilterBuilderUserEntry("ProductGroup", "Transactions_DM", dimFilterList)
     *
     * def dimFilterMap = ["CustomerId" : "CD-00155", "ProductId" : "MB-0005"]
     * def filter3 = datamartFilterBuilderUserEntry("ProductGroup", "Transactions_DM", dimFilterMap, Filter.equal("Country", "DE"))
     * </pre>
     * </blockquote>
     * @param paramName    Name of the user entry parameter.
     * @param source       Name of the source for which the filter is set.  The resolution of the source is as follows:
     * <ol>
     *                  <li> any Datamart - by typedId, sourceName, uniqueName or label (uniqueName and label for backwards compatibility) </li>
     *                  <li> any FieldCollection - by typedId or sourceName </li>
     * </ol>
     * @param args There can be up to 2 arguments: one can be a Map or List and the other a Filter. The order does not matter.
     *             If the argument is a Map, each pair [fieldName : value] will be used as a filter.
     * @return Datamart filter builder.
     * @see PublicGroovyAPI#datamartFilterBuilderUserEntry(String, String, Object...)
     * @see DmFilterBuilder
     */
    public DmFilterBuilder createDmFilter(String paramName, String fcTypedId, String field) {
        return new DmFilterBuilder(builderContext, paramName, fcTypedId, field);
    }


}
