package net.pricefx.server.dto.calculation;

import net.pricefx.common.api.CalculationResultType;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

public class ResultHighchart implements ComplexCalculationResult {

	private static final long serialVersionUID = 6101L;
	
	private final Map<String, ?> chartDefinition;
	
	private final List<String> modulesList = new ArrayList<>();
	
	private boolean showDataTab = false;
	
	private List<String> headersList = null;
	
	private List<ResultMatrix> resultMatrixList = null;
	
	public ResultHighchart(final Map<String, ?> definition) {
		this.chartDefinition = definition;
	}
	
	public Map<String, ?> getChartDefinition() {
		return chartDefinition;
	}

	/**
	 * By default, only the modules required for the various Price Analyzer chart types are available.
	 * If your chart requires additional modules, they can be specified here. The application ensures
	 * the files load only once.
	 * 
	 * @param modules    one or more Highcharts module names, without the path and without the ".js" extension.
	 *                   e.g. "annotations", "funnel", "variwide"
	 * @return the {@link ResultHighchart} instance, for convenience
	 * @see <a href="https://www.highcharts.com/docs">https://www.highcharts.com/docs</a> to find out what modules
	 *      are needed for each chart type.
	 */
	@SuppressWarnings("unchecked")
	public ResultHighchart addModule(String... modules) {
//		modulesList.addAll(Collections.arrayToList(modules));
		return this;
	}

	public List<String> getModulesList() {
		return modulesList;
	}
	
	/**
	 * Default is false.
	 * 
	 * Note that for convenience, calling setDataHeaders(), setResultMatrixTabs(), or
	 * addResultMatrixTab() will set this to true.
	 * 
	 * @param show    Data tab will be shown if set to true
	 * @return the {@link ResultHighchart} instance, for convenience
	 */
	public ResultHighchart setShowDataTab(final boolean show) {
		showDataTab = show;
		return this;
	}
	
	/**
	 * Short for setShowDataTab(true)
	 * @return the {@link ResultHighchart} instance, for convenience
	 */
	public ResultHighchart showDataTab() {
		return setShowDataTab(true);
	}
	
	public boolean getShowDataTab() {
		return showDataTab;
	}

	/**
	 * To override the default column headers of the data tab table.
	 * 
	 * For convenience, this will also call setShowDataTab(true)
	 * 
	 * @param headers    One or more header labels, left to right
	 * @return the {@link ResultHighchart} instance, for convenience
	 */
	@SuppressWarnings("unchecked")
	public ResultHighchart setDataHeaders(String... headers) {
//		headersList = Collections.arrayToList(headers);
		return showDataTab();
	}

	public List<String> getDataHeaders() {
		return headersList;
	}
	
	/**
	 * The automatically filled data tab be can be replaced by your own customized
	 * ResultMatrix data. It's even possible to add multiple tabs by providing multiple
	 * ResultMatrix instances.
	 * 
	 * Make sure to set a title via ResultMatrix.setTitle() as this will be used for the Tab label.
	 * 
	 * For convenience, this will also call setShowDataTab(true)
	 * 
	 * @param matrices    One or more ResultMatrix instances to display in the tab set
	 * @return the {@link ResultHighchart} instance, for convenience
	 */
	@SuppressWarnings("unchecked")
	public ResultHighchart setResultMatrixTabs(ResultMatrix... matrices) {
//		resultMatrixList = Collections.arrayToList(matrices);
		return showDataTab();
	}
	
	public ResultHighchart addResultMatrixTab(ResultMatrix matrix) {
		if (resultMatrixList == null) {
			resultMatrixList = new ArrayList<>();
		}
		resultMatrixList.add(matrix);
		return showDataTab();
	}
	
	public List<ResultMatrix> getResultMatrixList() {
		return resultMatrixList;
	}

	@Override
	public String getResultType() {
		return CalculationResultType.HIGHCHART.name();
	}

}
