package net.pricefx.service;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.googlecode.genericdao.search.Filter;


public class ItemSummaryQuery<T> {
	
	public static final String TOTAL_KEY = "TOTAL";
	
	public interface QueryGeneratorCallback {
		public String castNumeric(String fieldName);
		public String constantSelectTextExpression(String text);
	}
	
	public static class Projection {
		private String fieldNamePrefix;
		private String fieldName;
		private String aggregationMode;
		private String weight;
		private String distinctName;
		
		public Projection(String distinctName,String fieldName, String aggregationMode, String weight) {
			this(distinctName, "", fieldName, aggregationMode, weight);
		}
		
		public Projection(String distinctName, String fieldNamePrefix, String fieldName, String aggregationMode, String weight) {
			this.fieldNamePrefix = fieldNamePrefix;
			this.fieldName = fieldName;
			this.aggregationMode = aggregationMode;
			this.weight = weight;
			this.distinctName = distinctName;
		}
		
		public String getFieldNamePrefix() {
			return fieldNamePrefix;
		}
		
		public String getFieldName() {
			return fieldName;
		}
		
		public String getAggregationMode() {
			return aggregationMode;
		}

		public String getWeight() {
			return weight;
		}
		
		/** Used in SQL queries, overcomes ORACLE 32byte length identifier limit */
		public String getDistinctName() {
			return distinctName;
		}
		
		/** Client side distinct name. Match distinct names on client. */

		@JsonIgnore
		public String getClientSideDistinctName() {
			return String.format("%s%s_%s_%s", 
					fieldNamePrefix == null ? "" : cleanName(fieldNamePrefix),
					aggregationMode == null ? "" : aggregationMode,
					fieldName == null ? "" : cleanName(fieldName),
					weight == null ? "" : cleanName(weight)).toLowerCase();
		}
		
		private String cleanName(String fieldName) {
			if(fieldName != null) return fieldName.replace("%", "_x_").replace("*", "_x_").replace(" ", "");
			else return null;
		}		
		
	}

	public static class IDGenerator {

		private String prefix;
		private int idGenerator;
		
		public IDGenerator(String prefix) {
			this.prefix = prefix;
		}
		
		public String getNext() {
			return prefix +  idGenerator++;
		}

	}
	
	private String itemToObjectFkName;
	private String itemSqlAlias;
	private String itemGroupBy;
	private Filter itemFilter;
	private Filter productFilter;
	private String productGroupBy;
	private List<Projection> projections;
	private List<String> objectTypedIds;
	/* ID is unique inside query */
	private IDGenerator distinctNameGenerator;
	private boolean doCount;
	private QueryGeneratorCallback castCallback;
	
	public ItemSummaryQuery(String itemToObjectFkName) {
		this.itemToObjectFkName = itemToObjectFkName;
		itemSqlAlias = "cObjItem";
		
		// used for distinct names
		distinctNameGenerator = new IDGenerator("dn_");
	}
	
	@JsonIgnore
	public boolean isDualGroupBy() {
		return (itemGroupBy != null && !(itemGroupBy.isEmpty())) && (productGroupBy != null && !(productGroupBy.isEmpty()));
	}

	@JsonIgnore
	public boolean isNoGroupBy() {
		return (itemGroupBy == null || itemGroupBy.isEmpty()) && (productGroupBy == null || productGroupBy.isEmpty());
	}
	
	public String getItemGroupBy() {
		return itemGroupBy;
	}


	public void setItemGroupBy(String itemGroupBy) {
		this.itemGroupBy = itemGroupBy;
	}

	public boolean isDoCount() {
		return doCount;
	}

	public void setDoCount(boolean doCount) {
		this.doCount = doCount;
	}

	public String getItemToObjectFkName() {
		return itemToObjectFkName;
	}

	public String getItemSqlAlias() {
		return itemSqlAlias;
	}
	
	public void setItemSqlAlias(String itemSqlAlias) {
		this.itemSqlAlias = itemSqlAlias;
	}

	public Filter getItemFilter() {
		return itemFilter;
	}


	public void setItemFilter(Filter itemFilter) {
		this.itemFilter = itemFilter;
	}
	
	public void setCastCallback(QueryGeneratorCallback castCallback) {
		this.castCallback = castCallback;
	}

	public Filter getProductFilter() {
		return productFilter;
	}


	public void setProductFilter(Filter productFilter) {
		this.productFilter = productFilter;
	}


	public String getProductGroupBy() {
		return productGroupBy;
	}


	public void setProductGroupBy(String productGroupBy) {
		this.productGroupBy = productGroupBy;
	}

	public List<String> getObjectTypedIds() {
		return objectTypedIds;
	}

	public void setObjectTypedIds(List<String> objectTypedIds) {
		this.objectTypedIds = objectTypedIds;
	}
	
	public void addObjectTypedId(String objectTypedId) {
		if(objectTypedIds == null) objectTypedIds = new ArrayList<>();
		objectTypedIds.add(objectTypedId);
	}	

	public List<Projection> getProjections() {
		return projections;
	}
	public void setProjections(List<Projection> projections) {
		this.projections = projections;
	}

	@JsonIgnore
	public int getNumberOfProjections() {
		if(projections != null)
			return projections.size();
		
		return 0;
	}

	public void addProjection(String field, String aggregation) {
		addProjection(field, aggregation, null);
	}

	public void addProjection(String field, String aggregation, String weight) {
		addProjection("", field, aggregation, weight);
	}
	
	public void addProjection(String fieldPrefix, String field, String aggregation, String weight) {
		if(projections == null)
			projections = new ArrayList<>();
		
		projections.add(new Projection(distinctNameGenerator.getNext(),fieldPrefix,field,aggregation,weight));		
	}
	

	@SuppressWarnings("rawtypes")
	public void aggregate(Map<String,Object[]> results, List queryResults) {
		int noOfGroupBys = isDualGroupBy() ? 2 : 1;
		int countField = doCount ? 1 : 0;
		if(queryResults == null || results == null)
			return;
		
		for(int i=0; i<queryResults.size(); i++) {
			if(!(queryResults.get(i) instanceof Object[])) {
				break;
			}
			Object[] line = (Object[]) queryResults.get(i);
			String key = isDualGroupBy() ? String.format("%s_%s", line[0], line[1]) : String.format("%s", line[0]);
			
			if(results.containsKey(key)) {
				//we need to merge here
				
				//sanity check: the size of the result array should always be "noOfGroupBys + 2x" (i.e. 2 items per projection plus the groupBys
				if((line.length - noOfGroupBys - countField) % 2 != 0) {
				} else {
					aggregateRow(results.get(key),line);					
				}				
			} else {
				//just append as is
				results.put(key, line);
			}
		}		
	}
	
	public void addTotal(Map<String, Object[]> results) {
		if(results == null || results.size() == 0)
			return;
		
		if(results.containsKey(TOTAL_KEY))	//this happens if we do not have any group bys. Then the total comes straight from the DB already
			return;
		
		Object[] total = new Object[results.values().iterator().next().length];
		
		//init
		for(int i=0;i<total.length;i++) total[i] = new Double(0);
		
		//calc total
		for(String key : results.keySet()) {
			Object[] line = results.get(key);
			aggregateRow(total,line);
		}
				
		total[0] = TOTAL_KEY;
		if(isDualGroupBy()) total[1] = TOTAL_KEY;
		
		results.put(TOTAL_KEY, total);
		
	}

	private void aggregateRow(Object[] resultLine, Object[] line) {

		int noOfGroupBys = isDualGroupBy() ? 2 : 1;
		int countField = doCount ? 1 : 0;
		int iterations = (line.length - noOfGroupBys - countField) / 2;

		if(doCount) {
			// count field
			resultLine[noOfGroupBys] = aggregateValue((Number)resultLine[noOfGroupBys], (Number)line[noOfGroupBys], null, null);
		}

		for(int j=0; j< iterations; j++) {

			int x = noOfGroupBys + countField + j*2;

			// total
			resultLine[x] = aggregateValue((Number)resultLine[x],(Number)line[x],(Number)resultLine[x+1], (Number)line[x+1]);

			//new weight
			if(resultLine[x+1] != null && line[x+1] != null)
				resultLine[x+1] = ((Number)resultLine[x+1]).doubleValue() + ((Number)line[x+1]).doubleValue();
			else if (resultLine[x+1] == null && line[x+1] != null)
				resultLine[x+1] = line[x+1];

		}
	}

	private Number aggregateValue(Number oldVal, Number newVal, Number oldWeight, Number newWeight) {
		if(oldVal == null) return newVal;
		if(newVal == null) return oldVal;

		if((oldWeight == null || oldWeight.doubleValue() == 0) || (newWeight == null || newWeight.doubleValue() == 0))
			return oldVal.doubleValue() + newVal.doubleValue();
		else {
			return ((oldVal.doubleValue() * oldWeight.doubleValue()) + (newVal.doubleValue() * newWeight.doubleValue())) / (oldWeight.doubleValue() + newWeight.doubleValue());
		}
	}

	protected String getProjectionName(Projection p) {
		return p.getClientSideDistinctName();
	}
	
	public List<Map<String,Object>> asListOfMaps(Map<String, Object[]> resultList) {
		List<String> projectionNames = new ArrayList<>();
		projectionNames.add("groupKey");
		if(isDualGroupBy())
			projectionNames.add("groupKey2");
		if(isDoCount())
			projectionNames.add("countKey");
		
		for(Projection p : projections) {
			final String projName = getProjectionName(p);
			projectionNames.add(projName);
			projectionNames.add(String.format("totalWeight_%s", projName));	
		}
		
		List<Map<String,Object>> data = new ArrayList<>();
		
		if(resultList != null) {
			for(String key : resultList.keySet()) {
				Object[] resultRow = resultList.get(key);
				Map<String,Object> dataRow = new HashMap<>();
				for(int i=0;i<resultRow.length;i++) {
					dataRow.put(projectionNames.get(i),resultRow[i]);
				}
				postProcessRow(dataRow);
				data.add(dataRow);
			}	
		}
		return data;
	}
	
	protected void postProcessRow(Map<String, Object> row) {
	}

}
